import enum
import logging
import time
from contextlib import AbstractContextManager
from typing import Optional, TypeVar

from aperture_sdk._gen.aperture.flowcontrol.check.v1 import check_pb2
from aperture_sdk.const import (
    check_response_label,
    flow_end_timestamp_label,
    flow_status_label,
)
from google.protobuf import json_format
from opentelemetry import trace


class FlowDecision(enum.Enum):
    Accepted = enum.auto()
    Rejected = enum.auto()
    Unreachable = enum.auto()


class FlowStatus(enum.Enum):
    OK = enum.auto()
    Error = enum.auto()


TFlow = TypeVar("TFlow", bound="Flow")


class Flow(AbstractContextManager):
    def __init__(
        self, span: trace.Span, check_response: Optional[check_pb2.CheckResponse]
    ):
        self._span = span
        self._check_response = check_response
        self._status_code = FlowStatus.OK
        self._ended = False
        self._ramp_mode = False
        self.logger = logging.getLogger("aperture-py-sdk-flow")

    def should_run(self) -> bool:
        return self.decision == FlowDecision.Accepted or (
            (not self._ramp_mode) and self.decision == FlowDecision.Unreachable
        )

    @property
    def decision(self) -> FlowDecision:
        if self.check_response is None:
            return FlowDecision.Unreachable
        if (
            self.check_response.decision_type
            == check_pb2.CheckResponse.DECISION_TYPE_ACCEPTED
        ):
            return FlowDecision.Accepted
        return FlowDecision.Rejected

    @property
    def success(self) -> bool:
        return self.decision != FlowDecision.Unreachable

    @property
    def check_response(self) -> Optional[check_pb2.CheckResponse]:
        return self._check_response

    def set_status(self, status_code: FlowStatus) -> None:
        self._status_code = status_code

    def end(self) -> None:
        if self._ended:
            self.logger.warning("attempting to end an already ended flow")
            return
        self._ended = True

        check_response_json = (
            json_format.MessageToJson(self.check_response)
            if self.check_response
            else ""
        )
        self._span.set_attributes(
            {
                flow_status_label: self._status_code.name,
                check_response_label: check_response_json,
                flow_end_timestamp_label: time.monotonic_ns(),
            }
        )
        self._span.end()

    def __enter__(self: TFlow) -> TFlow:
        return self

    def __exit__(self, exc_type, _exc_value, _traceback) -> None:
        if self._ended:
            return
        if exc_type is not None:
            self.set_status(FlowStatus.Error)
        self.end()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import check_pb2 as aperture_dot_flowcontrol_dot_check_dot_v1_dot_check__pb2


class FlowControlServiceStub(object):
    """FlowControlService is used to perform Flow Control operations.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Check = channel.unary_unary(
                '/aperture.flowcontrol.check.v1.FlowControlService/Check',
                request_serializer=aperture_dot_flowcontrol_dot_check_dot_v1_dot_check__pb2.CheckRequest.SerializeToString,
                response_deserializer=aperture_dot_flowcontrol_dot_check_dot_v1_dot_check__pb2.CheckResponse.FromString,
                )


class FlowControlServiceServicer(object):
    """FlowControlService is used to perform Flow Control operations.
    """

    def Check(self, request, context):
        """Check wraps the given arbitrary resource and matches the given labels against Flow Control Limiters to makes a decision whether to allow/deny.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_FlowControlServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Check': grpc.unary_unary_rpc_method_handler(
                    servicer.Check,
                    request_deserializer=aperture_dot_flowcontrol_dot_check_dot_v1_dot_check__pb2.CheckRequest.FromString,
                    response_serializer=aperture_dot_flowcontrol_dot_check_dot_v1_dot_check__pb2.CheckResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'aperture.flowcontrol.check.v1.FlowControlService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class FlowControlService(object):
    """FlowControlService is used to perform Flow Control operations.
    """

    @staticmethod
    def Check(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/aperture.flowcontrol.check.v1.FlowControlService/Check',
            aperture_dot_flowcontrol_dot_check_dot_v1_dot_check__pb2.CheckRequest.SerializeToString,
            aperture_dot_flowcontrol_dot_check_dot_v1_dot_check__pb2.CheckResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: aperture/flowcontrol/check/v1/check.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n)aperture/flowcontrol/check/v1/check.proto\x12\x1d\x61perture.flowcontrol.check.v1\x1a\x1egoogle/protobuf/duration.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xdc\x01\n\x0c\x43heckRequest\x12#\n\rcontrol_point\x18\x01 \x01(\tR\x0c\x63ontrolPoint\x12O\n\x06labels\x18\x02 \x03(\x0b\x32\x37.aperture.flowcontrol.check.v1.CheckRequest.LabelsEntryR\x06labels\x12\x1b\n\tramp_mode\x18\x03 \x01(\x08R\x08rampMode\x1a\x39\n\x0bLabelsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xfd\t\n\rCheckResponse\x12\x30\n\x05start\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x05start\x12,\n\x03\x65nd\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x03\x65nd\x12\x1a\n\x08services\x18\x04 \x03(\tR\x08services\x12#\n\rcontrol_point\x18\x05 \x01(\tR\x0c\x63ontrolPoint\x12&\n\x0f\x66low_label_keys\x18\x06 \x03(\tR\rflowLabelKeys\x12y\n\x15telemetry_flow_labels\x18\x07 \x03(\x0b\x32\x45.aperture.flowcontrol.check.v1.CheckResponse.TelemetryFlowLabelsEntryR\x13telemetryFlowLabels\x12^\n\rdecision_type\x18\x08 \x01(\x0e\x32\x39.aperture.flowcontrol.check.v1.CheckResponse.DecisionTypeR\x0c\x64\x65\x63isionType\x12^\n\rreject_reason\x18\t \x01(\x0e\x32\x39.aperture.flowcontrol.check.v1.CheckResponse.RejectReasonR\x0crejectReason\x12X\n\x10\x63lassifier_infos\x18\n \x03(\x0b\x32-.aperture.flowcontrol.check.v1.ClassifierInfoR\x0f\x63lassifierInfos\x12V\n\x10\x66lux_meter_infos\x18\x0b \x03(\x0b\x32,.aperture.flowcontrol.check.v1.FluxMeterInfoR\x0e\x66luxMeterInfos\x12[\n\x11limiter_decisions\x18\x0c \x03(\x0b\x32..aperture.flowcontrol.check.v1.LimiterDecisionR\x10limiterDecisions\x12\x36\n\twait_time\x18\r \x01(\x0b\x32\x19.google.protobuf.DurationR\x08waitTime\x12h\n\x1b\x64\x65nied_response_status_code\x18\x0e \x01(\x0e\x32).aperture.flowcontrol.check.v1.StatusCodeR\x18\x64\x65niedResponseStatusCode\x1a\x46\n\x18TelemetryFlowLabelsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\xa6\x01\n\x0cRejectReason\x12\x16\n\x12REJECT_REASON_NONE\x10\x00\x12\x1e\n\x1aREJECT_REASON_RATE_LIMITED\x10\x01\x12\x1b\n\x17REJECT_REASON_NO_TOKENS\x10\x02\x12\x1d\n\x19REJECT_REASON_NOT_SAMPLED\x10\x03\x12\"\n\x1eREJECT_REASON_NO_MATCHING_RAMP\x10\x04\"F\n\x0c\x44\x65\x63isionType\x12\x1a\n\x16\x44\x45\x43ISION_TYPE_ACCEPTED\x10\x00\x12\x1a\n\x16\x44\x45\x43ISION_TYPE_REJECTED\x10\x01\"\xed\x02\n\x0e\x43lassifierInfo\x12\x1f\n\x0bpolicy_name\x18\x01 \x01(\tR\npolicyName\x12\x1f\n\x0bpolicy_hash\x18\x02 \x01(\tR\npolicyHash\x12)\n\x10\x63lassifier_index\x18\x03 \x01(\x03R\x0f\x63lassifierIndex\x12I\n\x05\x65rror\x18\x05 \x01(\x0e\x32\x33.aperture.flowcontrol.check.v1.ClassifierInfo.ErrorR\x05\x65rror\"\xa2\x01\n\x05\x45rror\x12\x0e\n\nERROR_NONE\x10\x00\x12\x15\n\x11\x45RROR_EVAL_FAILED\x10\x01\x12\x19\n\x15\x45RROR_EMPTY_RESULTSET\x10\x02\x12\x1d\n\x19\x45RROR_AMBIGUOUS_RESULTSET\x10\x03\x12\x1a\n\x16\x45RROR_MULTI_EXPRESSION\x10\x04\x12\x1c\n\x18\x45RROR_EXPRESSION_NOT_MAP\x10\x05\"\xa6\x0c\n\x0fLimiterDecision\x12\x1f\n\x0bpolicy_name\x18\x01 \x01(\tR\npolicyName\x12\x1f\n\x0bpolicy_hash\x18\x02 \x01(\tR\npolicyHash\x12!\n\x0c\x63omponent_id\x18\x03 \x01(\tR\x0b\x63omponentId\x12\x18\n\x07\x64ropped\x18\x04 \x01(\x08R\x07\x64ropped\x12T\n\x06reason\x18\x05 \x01(\x0e\x32<.aperture.flowcontrol.check.v1.LimiterDecision.LimiterReasonR\x06reason\x12h\n\x1b\x64\x65nied_response_status_code\x18\n \x01(\x0e\x32).aperture.flowcontrol.check.v1.StatusCodeR\x18\x64\x65niedResponseStatusCode\x12\x36\n\twait_time\x18\x0b \x01(\x0b\x32\x19.google.protobuf.DurationR\x08waitTime\x12l\n\x11rate_limiter_info\x18\x14 \x01(\x0b\x32>.aperture.flowcontrol.check.v1.LimiterDecision.RateLimiterInfoH\x00R\x0frateLimiterInfo\x12n\n\x13load_scheduler_info\x18\x15 \x01(\x0b\x32<.aperture.flowcontrol.check.v1.LimiterDecision.SchedulerInfoH\x00R\x11loadSchedulerInfo\x12_\n\x0csampler_info\x18\x16 \x01(\x0b\x32:.aperture.flowcontrol.check.v1.LimiterDecision.SamplerInfoH\x00R\x0bsamplerInfo\x12u\n\x14quota_scheduler_info\x18\x17 \x01(\x0b\x32\x41.aperture.flowcontrol.check.v1.LimiterDecision.QuotaSchedulerInfoH\x00R\x12quotaSchedulerInfo\x1a`\n\nTokensInfo\x12\x1c\n\tremaining\x18\x01 \x01(\x01R\tremaining\x12\x18\n\x07\x63urrent\x18\x02 \x01(\x01R\x07\x63urrent\x12\x1a\n\x08\x63onsumed\x18\x03 \x01(\x01R\x08\x63onsumed\x1a\x83\x01\n\x0fRateLimiterInfo\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x12Z\n\x0btokens_info\x18\x02 \x01(\x0b\x32\x39.aperture.flowcontrol.check.v1.LimiterDecision.TokensInfoR\ntokensInfo\x1a\xae\x01\n\rSchedulerInfo\x12%\n\x0eworkload_index\x18\x01 \x01(\tR\rworkloadIndex\x12Z\n\x0btokens_info\x18\x02 \x01(\x0b\x32\x39.aperture.flowcontrol.check.v1.LimiterDecision.TokensInfoR\ntokensInfo\x12\x1a\n\x08priority\x18\x03 \x01(\x01R\x08priority\x1a#\n\x0bSamplerInfo\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x1a\xc9\x01\n\x12QuotaSchedulerInfo\x12\x14\n\x05label\x18\x01 \x01(\tR\x05label\x12%\n\x0eworkload_index\x18\x02 \x01(\tR\rworkloadIndex\x12Z\n\x0btokens_info\x18\x03 \x01(\x0b\x32\x39.aperture.flowcontrol.check.v1.LimiterDecision.TokensInfoR\ntokensInfo\x12\x1a\n\x08priority\x18\x04 \x01(\x01R\x08priority\"Q\n\rLimiterReason\x12\x1e\n\x1aLIMITER_REASON_UNSPECIFIED\x10\x00\x12 \n\x1cLIMITER_REASON_KEY_NOT_FOUND\x10\x01\x42\t\n\x07\x64\x65tails\"7\n\rFluxMeterInfo\x12&\n\x0f\x66lux_meter_name\x18\x01 \x01(\tR\rfluxMeterName*\xb5\t\n\nStatusCode\x12\t\n\x05\x45mpty\x10\x00\x12\x0c\n\x08\x43ontinue\x10\x64\x12\x07\n\x02OK\x10\xc8\x01\x12\x0c\n\x07\x43reated\x10\xc9\x01\x12\r\n\x08\x41\x63\x63\x65pted\x10\xca\x01\x12 \n\x1bNonAuthoritativeInformation\x10\xcb\x01\x12\x0e\n\tNoContent\x10\xcc\x01\x12\x11\n\x0cResetContent\x10\xcd\x01\x12\x13\n\x0ePartialContent\x10\xce\x01\x12\x10\n\x0bMultiStatus\x10\xcf\x01\x12\x14\n\x0f\x41lreadyReported\x10\xd0\x01\x12\x0b\n\x06IMUsed\x10\xe2\x01\x12\x14\n\x0fMultipleChoices\x10\xac\x02\x12\x15\n\x10MovedPermanently\x10\xad\x02\x12\n\n\x05\x46ound\x10\xae\x02\x12\r\n\x08SeeOther\x10\xaf\x02\x12\x10\n\x0bNotModified\x10\xb0\x02\x12\r\n\x08UseProxy\x10\xb1\x02\x12\x16\n\x11TemporaryRedirect\x10\xb3\x02\x12\x16\n\x11PermanentRedirect\x10\xb4\x02\x12\x0f\n\nBadRequest\x10\x90\x03\x12\x11\n\x0cUnauthorized\x10\x91\x03\x12\x14\n\x0fPaymentRequired\x10\x92\x03\x12\x0e\n\tForbidden\x10\x93\x03\x12\r\n\x08NotFound\x10\x94\x03\x12\x15\n\x10MethodNotAllowed\x10\x95\x03\x12\x12\n\rNotAcceptable\x10\x96\x03\x12 \n\x1bProxyAuthenticationRequired\x10\x97\x03\x12\x13\n\x0eRequestTimeout\x10\x98\x03\x12\r\n\x08\x43onflict\x10\x99\x03\x12\t\n\x04Gone\x10\x9a\x03\x12\x13\n\x0eLengthRequired\x10\x9b\x03\x12\x17\n\x12PreconditionFailed\x10\x9c\x03\x12\x14\n\x0fPayloadTooLarge\x10\x9d\x03\x12\x0f\n\nURITooLong\x10\x9e\x03\x12\x19\n\x14UnsupportedMediaType\x10\x9f\x03\x12\x18\n\x13RangeNotSatisfiable\x10\xa0\x03\x12\x16\n\x11\x45xpectationFailed\x10\xa1\x03\x12\x17\n\x12MisdirectedRequest\x10\xa5\x03\x12\x18\n\x13UnprocessableEntity\x10\xa6\x03\x12\x0b\n\x06Locked\x10\xa7\x03\x12\x15\n\x10\x46\x61iledDependency\x10\xa8\x03\x12\x14\n\x0fUpgradeRequired\x10\xaa\x03\x12\x19\n\x14PreconditionRequired\x10\xac\x03\x12\x14\n\x0fTooManyRequests\x10\xad\x03\x12 \n\x1bRequestHeaderFieldsTooLarge\x10\xaf\x03\x12\x18\n\x13InternalServerError\x10\xf4\x03\x12\x13\n\x0eNotImplemented\x10\xf5\x03\x12\x0f\n\nBadGateway\x10\xf6\x03\x12\x17\n\x12ServiceUnavailable\x10\xf7\x03\x12\x13\n\x0eGatewayTimeout\x10\xf8\x03\x12\x1c\n\x17HTTPVersionNotSupported\x10\xf9\x03\x12\x1a\n\x15VariantAlsoNegotiates\x10\xfa\x03\x12\x18\n\x13InsufficientStorage\x10\xfb\x03\x12\x11\n\x0cLoopDetected\x10\xfc\x03\x12\x10\n\x0bNotExtended\x10\xfe\x03\x12\"\n\x1dNetworkAuthenticationRequired\x10\xff\x03\x32z\n\x12\x46lowControlService\x12\x64\n\x05\x43heck\x12+.aperture.flowcontrol.check.v1.CheckRequest\x1a,.aperture.flowcontrol.check.v1.CheckResponse\"\x00\x42\xb3\x02\n5com.fluxninja.generated.aperture.flowcontrol.check.v1B\nCheckProtoP\x01ZWgithub.com/fluxninja/aperture/v2/api/gen/proto/go/aperture/flowcontrol/check/v1;checkv1\xa2\x02\x03\x41\x46\x43\xaa\x02\x1d\x41perture.Flowcontrol.Check.V1\xca\x02\x1d\x41perture\\Flowcontrol\\Check\\V1\xe2\x02)Aperture\\Flowcontrol\\Check\\V1\\GPBMetadata\xea\x02 Aperture::Flowcontrol::Check::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'aperture.flowcontrol.check.v1.check_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n5com.fluxninja.generated.aperture.flowcontrol.check.v1B\nCheckProtoP\001ZWgithub.com/fluxninja/aperture/v2/api/gen/proto/go/aperture/flowcontrol/check/v1;checkv1\242\002\003AFC\252\002\035Aperture.Flowcontrol.Check.V1\312\002\035Aperture\\Flowcontrol\\Check\\V1\342\002)Aperture\\Flowcontrol\\Check\\V1\\GPBMetadata\352\002 Aperture::Flowcontrol::Check::V1'
  _globals['_CHECKREQUEST_LABELSENTRY']._options = None
  _globals['_CHECKREQUEST_LABELSENTRY']._serialized_options = b'8\001'
  _globals['_CHECKRESPONSE_TELEMETRYFLOWLABELSENTRY']._options = None
  _globals['_CHECKRESPONSE_TELEMETRYFLOWLABELSENTRY']._serialized_options = b'8\001'
  _globals['_STATUSCODE']._serialized_start=3647
  _globals['_STATUSCODE']._serialized_end=4852
  _globals['_CHECKREQUEST']._serialized_start=142
  _globals['_CHECKREQUEST']._serialized_end=362
  _globals['_CHECKREQUEST_LABELSENTRY']._serialized_start=305
  _globals['_CHECKREQUEST_LABELSENTRY']._serialized_end=362
  _globals['_CHECKRESPONSE']._serialized_start=365
  _globals['_CHECKRESPONSE']._serialized_end=1642
  _globals['_CHECKRESPONSE_TELEMETRYFLOWLABELSENTRY']._serialized_start=1331
  _globals['_CHECKRESPONSE_TELEMETRYFLOWLABELSENTRY']._serialized_end=1401
  _globals['_CHECKRESPONSE_REJECTREASON']._serialized_start=1404
  _globals['_CHECKRESPONSE_REJECTREASON']._serialized_end=1570
  _globals['_CHECKRESPONSE_DECISIONTYPE']._serialized_start=1572
  _globals['_CHECKRESPONSE_DECISIONTYPE']._serialized_end=1642
  _globals['_CLASSIFIERINFO']._serialized_start=1645
  _globals['_CLASSIFIERINFO']._serialized_end=2010
  _globals['_CLASSIFIERINFO_ERROR']._serialized_start=1848
  _globals['_CLASSIFIERINFO_ERROR']._serialized_end=2010
  _globals['_LIMITERDECISION']._serialized_start=2013
  _globals['_LIMITERDECISION']._serialized_end=3587
  _globals['_LIMITERDECISION_TOKENSINFO']._serialized_start=2845
  _globals['_LIMITERDECISION_TOKENSINFO']._serialized_end=2941
  _globals['_LIMITERDECISION_RATELIMITERINFO']._serialized_start=2944
  _globals['_LIMITERDECISION_RATELIMITERINFO']._serialized_end=3075
  _globals['_LIMITERDECISION_SCHEDULERINFO']._serialized_start=3078
  _globals['_LIMITERDECISION_SCHEDULERINFO']._serialized_end=3252
  _globals['_LIMITERDECISION_SAMPLERINFO']._serialized_start=3254
  _globals['_LIMITERDECISION_SAMPLERINFO']._serialized_end=3289
  _globals['_LIMITERDECISION_QUOTASCHEDULERINFO']._serialized_start=3292
  _globals['_LIMITERDECISION_QUOTASCHEDULERINFO']._serialized_end=3493
  _globals['_LIMITERDECISION_LIMITERREASON']._serialized_start=3495
  _globals['_LIMITERDECISION_LIMITERREASON']._serialized_end=3576
  _globals['_FLUXMETERINFO']._serialized_start=3589
  _globals['_FLUXMETERINFO']._serialized_end=3644
  _globals['_FLOWCONTROLSERVICE']._serialized_start=4854
  _globals['_FLOWCONTROLSERVICE']._serialized_end=4976
# @@protoc_insertion_point(module_scope)



from datetime import timedelta

# Library name and version can be used by the user to create a resource that connects to telemetry exporter.
library_name = "aperture-py"
library_version = "v2.4.0"

# Config defaults.
default_rpc_timeout = timedelta(milliseconds=200)
default_grpc_reconnection_time = timedelta(seconds=10)

# Label keys.
# Label to hold source of flow.
source_label = "aperture.source"
# Label to hold status of the flow.
flow_status_label = "aperture.flow.status"
# Label to hold JSON encoded check response struct.
check_response_label = "aperture.check_response"
# Label to hold flow's start timestamp in Unix nanoseconds since Epoch.
flow_start_timestamp_label = "aperture.flow_start_timestamp"
# Label to hold flow's stop timestamp in Unix nanoseconds since Epoch.
flow_end_timestamp_label = "aperture.flow_end_timestamp"
# Label to hold workload start timestamp in Unix nanoseconds since Epoch.
workload_start_timestamp_label = "aperture.workload_start_timestamp"



import inspect
from typing import Callable, TypeVar

TWrappedReturn = TypeVar("TWrappedReturn")


async def run_fn(fn: Callable[..., TWrappedReturn], *args, **kwargs) -> TWrappedReturn:
    """Run a function or coroutine."""
    # We want to support both sync and async functions
    # https://stackoverflow.com/questions/44169998/how-to-create-a-python-decorator-that-can-wrap-either-coroutine-or-function
    if inspect.iscoroutinefunction(fn):
        return await fn(*args, **kwargs)
    else:
        return fn(*args, **kwargs)



from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class StatusCode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    Empty: _ClassVar[StatusCode]
    Continue: _ClassVar[StatusCode]
    OK: _ClassVar[StatusCode]
    Created: _ClassVar[StatusCode]
    Accepted: _ClassVar[StatusCode]
    NonAuthoritativeInformation: _ClassVar[StatusCode]
    NoContent: _ClassVar[StatusCode]
    ResetContent: _ClassVar[StatusCode]
    PartialContent: _ClassVar[StatusCode]
    MultiStatus: _ClassVar[StatusCode]
    AlreadyReported: _ClassVar[StatusCode]
    IMUsed: _ClassVar[StatusCode]
    MultipleChoices: _ClassVar[StatusCode]
    MovedPermanently: _ClassVar[StatusCode]
    Found: _ClassVar[StatusCode]
    SeeOther: _ClassVar[StatusCode]
    NotModified: _ClassVar[StatusCode]
    UseProxy: _ClassVar[StatusCode]
    TemporaryRedirect: _ClassVar[StatusCode]
    PermanentRedirect: _ClassVar[StatusCode]
    BadRequest: _ClassVar[StatusCode]
    Unauthorized: _ClassVar[StatusCode]
    PaymentRequired: _ClassVar[StatusCode]
    Forbidden: _ClassVar[StatusCode]
    NotFound: _ClassVar[StatusCode]
    MethodNotAllowed: _ClassVar[StatusCode]
    NotAcceptable: _ClassVar[StatusCode]
    ProxyAuthenticationRequired: _ClassVar[StatusCode]
    RequestTimeout: _ClassVar[StatusCode]
    Conflict: _ClassVar[StatusCode]
    Gone: _ClassVar[StatusCode]
    LengthRequired: _ClassVar[StatusCode]
    PreconditionFailed: _ClassVar[StatusCode]
    PayloadTooLarge: _ClassVar[StatusCode]
    URITooLong: _ClassVar[StatusCode]
    UnsupportedMediaType: _ClassVar[StatusCode]
    RangeNotSatisfiable: _ClassVar[StatusCode]
    ExpectationFailed: _ClassVar[StatusCode]
    MisdirectedRequest: _ClassVar[StatusCode]
    UnprocessableEntity: _ClassVar[StatusCode]
    Locked: _ClassVar[StatusCode]
    FailedDependency: _ClassVar[StatusCode]
    UpgradeRequired: _ClassVar[StatusCode]
    PreconditionRequired: _ClassVar[StatusCode]
    TooManyRequests: _ClassVar[StatusCode]
    RequestHeaderFieldsTooLarge: _ClassVar[StatusCode]
    InternalServerError: _ClassVar[StatusCode]
    NotImplemented: _ClassVar[StatusCode]
    BadGateway: _ClassVar[StatusCode]
    ServiceUnavailable: _ClassVar[StatusCode]
    GatewayTimeout: _ClassVar[StatusCode]
    HTTPVersionNotSupported: _ClassVar[StatusCode]
    VariantAlsoNegotiates: _ClassVar[StatusCode]
    InsufficientStorage: _ClassVar[StatusCode]
    LoopDetected: _ClassVar[StatusCode]
    NotExtended: _ClassVar[StatusCode]
    NetworkAuthenticationRequired: _ClassVar[StatusCode]
Empty: StatusCode
Continue: StatusCode
OK: StatusCode
Created: StatusCode
Accepted: StatusCode
NonAuthoritativeInformation: StatusCode
NoContent: StatusCode
ResetContent: StatusCode
PartialContent: StatusCode
MultiStatus: StatusCode
AlreadyReported: StatusCode
IMUsed: StatusCode
MultipleChoices: StatusCode
MovedPermanently: StatusCode
Found: StatusCode
SeeOther: StatusCode
NotModified: StatusCode
UseProxy: StatusCode
TemporaryRedirect: StatusCode
PermanentRedirect: StatusCode
BadRequest: StatusCode
Unauthorized: StatusCode
PaymentRequired: StatusCode
Forbidden: StatusCode
NotFound: StatusCode
MethodNotAllowed: StatusCode
NotAcceptable: StatusCode
ProxyAuthenticationRequired: StatusCode
RequestTimeout: StatusCode
Conflict: StatusCode
Gone: StatusCode
LengthRequired: StatusCode
PreconditionFailed: StatusCode
PayloadTooLarge: StatusCode
URITooLong: StatusCode
UnsupportedMediaType: StatusCode
RangeNotSatisfiable: StatusCode
ExpectationFailed: StatusCode
MisdirectedRequest: StatusCode
UnprocessableEntity: StatusCode
Locked: StatusCode
FailedDependency: StatusCode
UpgradeRequired: StatusCode
PreconditionRequired: StatusCode
TooManyRequests: StatusCode
RequestHeaderFieldsTooLarge: StatusCode
InternalServerError: StatusCode
NotImplemented: StatusCode
BadGateway: StatusCode
ServiceUnavailable: StatusCode
GatewayTimeout: StatusCode
HTTPVersionNotSupported: StatusCode
VariantAlsoNegotiates: StatusCode
InsufficientStorage: StatusCode
LoopDetected: StatusCode
NotExtended: StatusCode
NetworkAuthenticationRequired: StatusCode

class CheckRequest(_message.Message):
    __slots__ = ["control_point", "labels", "ramp_mode"]
    class LabelsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    CONTROL_POINT_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    RAMP_MODE_FIELD_NUMBER: _ClassVar[int]
    control_point: str
    labels: _containers.ScalarMap[str, str]
    ramp_mode: bool
    def __init__(self, control_point: _Optional[str] = ..., labels: _Optional[_Mapping[str, str]] = ..., ramp_mode: bool = ...) -> None: ...

class CheckResponse(_message.Message):
    __slots__ = ["start", "end", "services", "control_point", "flow_label_keys", "telemetry_flow_labels", "decision_type", "reject_reason", "classifier_infos", "flux_meter_infos", "limiter_decisions", "wait_time", "denied_response_status_code"]
    class RejectReason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        REJECT_REASON_NONE: _ClassVar[CheckResponse.RejectReason]
        REJECT_REASON_RATE_LIMITED: _ClassVar[CheckResponse.RejectReason]
        REJECT_REASON_NO_TOKENS: _ClassVar[CheckResponse.RejectReason]
        REJECT_REASON_NOT_SAMPLED: _ClassVar[CheckResponse.RejectReason]
        REJECT_REASON_NO_MATCHING_RAMP: _ClassVar[CheckResponse.RejectReason]
    REJECT_REASON_NONE: CheckResponse.RejectReason
    REJECT_REASON_RATE_LIMITED: CheckResponse.RejectReason
    REJECT_REASON_NO_TOKENS: CheckResponse.RejectReason
    REJECT_REASON_NOT_SAMPLED: CheckResponse.RejectReason
    REJECT_REASON_NO_MATCHING_RAMP: CheckResponse.RejectReason
    class DecisionType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        DECISION_TYPE_ACCEPTED: _ClassVar[CheckResponse.DecisionType]
        DECISION_TYPE_REJECTED: _ClassVar[CheckResponse.DecisionType]
    DECISION_TYPE_ACCEPTED: CheckResponse.DecisionType
    DECISION_TYPE_REJECTED: CheckResponse.DecisionType
    class TelemetryFlowLabelsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    START_FIELD_NUMBER: _ClassVar[int]
    END_FIELD_NUMBER: _ClassVar[int]
    SERVICES_FIELD_NUMBER: _ClassVar[int]
    CONTROL_POINT_FIELD_NUMBER: _ClassVar[int]
    FLOW_LABEL_KEYS_FIELD_NUMBER: _ClassVar[int]
    TELEMETRY_FLOW_LABELS_FIELD_NUMBER: _ClassVar[int]
    DECISION_TYPE_FIELD_NUMBER: _ClassVar[int]
    REJECT_REASON_FIELD_NUMBER: _ClassVar[int]
    CLASSIFIER_INFOS_FIELD_NUMBER: _ClassVar[int]
    FLUX_METER_INFOS_FIELD_NUMBER: _ClassVar[int]
    LIMITER_DECISIONS_FIELD_NUMBER: _ClassVar[int]
    WAIT_TIME_FIELD_NUMBER: _ClassVar[int]
    DENIED_RESPONSE_STATUS_CODE_FIELD_NUMBER: _ClassVar[int]
    start: _timestamp_pb2.Timestamp
    end: _timestamp_pb2.Timestamp
    services: _containers.RepeatedScalarFieldContainer[str]
    control_point: str
    flow_label_keys: _containers.RepeatedScalarFieldContainer[str]
    telemetry_flow_labels: _containers.ScalarMap[str, str]
    decision_type: CheckResponse.DecisionType
    reject_reason: CheckResponse.RejectReason
    classifier_infos: _containers.RepeatedCompositeFieldContainer[ClassifierInfo]
    flux_meter_infos: _containers.RepeatedCompositeFieldContainer[FluxMeterInfo]
    limiter_decisions: _containers.RepeatedCompositeFieldContainer[LimiterDecision]
    wait_time: _duration_pb2.Duration
    denied_response_status_code: StatusCode
    def __init__(self, start: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., end: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., services: _Optional[_Iterable[str]] = ..., control_point: _Optional[str] = ..., flow_label_keys: _Optional[_Iterable[str]] = ..., telemetry_flow_labels: _Optional[_Mapping[str, str]] = ..., decision_type: _Optional[_Union[CheckResponse.DecisionType, str]] = ..., reject_reason: _Optional[_Union[CheckResponse.RejectReason, str]] = ..., classifier_infos: _Optional[_Iterable[_Union[ClassifierInfo, _Mapping]]] = ..., flux_meter_infos: _Optional[_Iterable[_Union[FluxMeterInfo, _Mapping]]] = ..., limiter_decisions: _Optional[_Iterable[_Union[LimiterDecision, _Mapping]]] = ..., wait_time: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., denied_response_status_code: _Optional[_Union[StatusCode, str]] = ...) -> None: ...

class ClassifierInfo(_message.Message):
    __slots__ = ["policy_name", "policy_hash", "classifier_index", "error"]
    class Error(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        ERROR_NONE: _ClassVar[ClassifierInfo.Error]
        ERROR_EVAL_FAILED: _ClassVar[ClassifierInfo.Error]
        ERROR_EMPTY_RESULTSET: _ClassVar[ClassifierInfo.Error]
        ERROR_AMBIGUOUS_RESULTSET: _ClassVar[ClassifierInfo.Error]
        ERROR_MULTI_EXPRESSION: _ClassVar[ClassifierInfo.Error]
        ERROR_EXPRESSION_NOT_MAP: _ClassVar[ClassifierInfo.Error]
    ERROR_NONE: ClassifierInfo.Error
    ERROR_EVAL_FAILED: ClassifierInfo.Error
    ERROR_EMPTY_RESULTSET: ClassifierInfo.Error
    ERROR_AMBIGUOUS_RESULTSET: ClassifierInfo.Error
    ERROR_MULTI_EXPRESSION: ClassifierInfo.Error
    ERROR_EXPRESSION_NOT_MAP: ClassifierInfo.Error
    POLICY_NAME_FIELD_NUMBER: _ClassVar[int]
    POLICY_HASH_FIELD_NUMBER: _ClassVar[int]
    CLASSIFIER_INDEX_FIELD_NUMBER: _ClassVar[int]
    ERROR_FIELD_NUMBER: _ClassVar[int]
    policy_name: str
    policy_hash: str
    classifier_index: int
    error: ClassifierInfo.Error
    def __init__(self, policy_name: _Optional[str] = ..., policy_hash: _Optional[str] = ..., classifier_index: _Optional[int] = ..., error: _Optional[_Union[ClassifierInfo.Error, str]] = ...) -> None: ...

class LimiterDecision(_message.Message):
    __slots__ = ["policy_name", "policy_hash", "component_id", "dropped", "reason", "denied_response_status_code", "wait_time", "rate_limiter_info", "load_scheduler_info", "sampler_info", "quota_scheduler_info"]
    class LimiterReason(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        LIMITER_REASON_UNSPECIFIED: _ClassVar[LimiterDecision.LimiterReason]
        LIMITER_REASON_KEY_NOT_FOUND: _ClassVar[LimiterDecision.LimiterReason]
    LIMITER_REASON_UNSPECIFIED: LimiterDecision.LimiterReason
    LIMITER_REASON_KEY_NOT_FOUND: LimiterDecision.LimiterReason
    class TokensInfo(_message.Message):
        __slots__ = ["remaining", "current", "consumed"]
        REMAINING_FIELD_NUMBER: _ClassVar[int]
        CURRENT_FIELD_NUMBER: _ClassVar[int]
        CONSUMED_FIELD_NUMBER: _ClassVar[int]
        remaining: float
        current: float
        consumed: float
        def __init__(self, remaining: _Optional[float] = ..., current: _Optional[float] = ..., consumed: _Optional[float] = ...) -> None: ...
    class RateLimiterInfo(_message.Message):
        __slots__ = ["label", "tokens_info"]
        LABEL_FIELD_NUMBER: _ClassVar[int]
        TOKENS_INFO_FIELD_NUMBER: _ClassVar[int]
        label: str
        tokens_info: LimiterDecision.TokensInfo
        def __init__(self, label: _Optional[str] = ..., tokens_info: _Optional[_Union[LimiterDecision.TokensInfo, _Mapping]] = ...) -> None: ...
    class SchedulerInfo(_message.Message):
        __slots__ = ["workload_index", "tokens_info", "priority"]
        WORKLOAD_INDEX_FIELD_NUMBER: _ClassVar[int]
        TOKENS_INFO_FIELD_NUMBER: _ClassVar[int]
        PRIORITY_FIELD_NUMBER: _ClassVar[int]
        workload_index: str
        tokens_info: LimiterDecision.TokensInfo
        priority: float
        def __init__(self, workload_index: _Optional[str] = ..., tokens_info: _Optional[_Union[LimiterDecision.TokensInfo, _Mapping]] = ..., priority: _Optional[float] = ...) -> None: ...
    class SamplerInfo(_message.Message):
        __slots__ = ["label"]
        LABEL_FIELD_NUMBER: _ClassVar[int]
        label: str
        def __init__(self, label: _Optional[str] = ...) -> None: ...
    class QuotaSchedulerInfo(_message.Message):
        __slots__ = ["label", "workload_index", "tokens_info", "priority"]
        LABEL_FIELD_NUMBER: _ClassVar[int]
        WORKLOAD_INDEX_FIELD_NUMBER: _ClassVar[int]
        TOKENS_INFO_FIELD_NUMBER: _ClassVar[int]
        PRIORITY_FIELD_NUMBER: _ClassVar[int]
        label: str
        workload_index: str
        tokens_info: LimiterDecision.TokensInfo
        priority: float
        def __init__(self, label: _Optional[str] = ..., workload_index: _Optional[str] = ..., tokens_info: _Optional[_Union[LimiterDecision.TokensInfo, _Mapping]] = ..., priority: _Optional[float] = ...) -> None: ...
    POLICY_NAME_FIELD_NUMBER: _ClassVar[int]
    POLICY_HASH_FIELD_NUMBER: _ClassVar[int]
    COMPONENT_ID_FIELD_NUMBER: _ClassVar[int]
    DROPPED_FIELD_NUMBER: _ClassVar[int]
    REASON_FIELD_NUMBER: _ClassVar[int]
    DENIED_RESPONSE_STATUS_CODE_FIELD_NUMBER: _ClassVar[int]
    WAIT_TIME_FIELD_NUMBER: _ClassVar[int]
    RATE_LIMITER_INFO_FIELD_NUMBER: _ClassVar[int]
    LOAD_SCHEDULER_INFO_FIELD_NUMBER: _ClassVar[int]
    SAMPLER_INFO_FIELD_NUMBER: _ClassVar[int]
    QUOTA_SCHEDULER_INFO_FIELD_NUMBER: _ClassVar[int]
    policy_name: str
    policy_hash: str
    component_id: str
    dropped: bool
    reason: LimiterDecision.LimiterReason
    denied_response_status_code: StatusCode
    wait_time: _duration_pb2.Duration
    rate_limiter_info: LimiterDecision.RateLimiterInfo
    load_scheduler_info: LimiterDecision.SchedulerInfo
    sampler_info: LimiterDecision.SamplerInfo
    quota_scheduler_info: LimiterDecision.QuotaSchedulerInfo
    def __init__(self, policy_name: _Optional[str] = ..., policy_hash: _Optional[str] = ..., component_id: _Optional[str] = ..., dropped: bool = ..., reason: _Optional[_Union[LimiterDecision.LimiterReason, str]] = ..., denied_response_status_code: _Optional[_Union[StatusCode, str]] = ..., wait_time: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., rate_limiter_info: _Optional[_Union[LimiterDecision.RateLimiterInfo, _Mapping]] = ..., load_scheduler_info: _Optional[_Union[LimiterDecision.SchedulerInfo, _Mapping]] = ..., sampler_info: _Optional[_Union[LimiterDecision.SamplerInfo, _Mapping]] = ..., quota_scheduler_info: _Optional[_Union[LimiterDecision.QuotaSchedulerInfo, _Mapping]] = ...) -> None: ...

class FluxMeterInfo(_message.Message):
    __slots__ = ["flux_meter_name"]
    FLUX_METER_NAME_FIELD_NUMBER: _ClassVar[int]
    flux_meter_name: str
    def __init__(self, flux_meter_name: _Optional[str] = ...) -> None: ...



import datetime
import functools
import logging
import time
import typing
from typing import Callable, Dict, Optional, Type, TypeVar

import grpc
from aperture_sdk._gen.aperture.flowcontrol.check.v1.check_pb2 import CheckRequest
from aperture_sdk._gen.aperture.flowcontrol.check.v1.check_pb2_grpc import (
    FlowControlServiceStub,
)
from aperture_sdk.const import (
    default_grpc_reconnection_time,
    default_rpc_timeout,
    flow_start_timestamp_label,
    library_name,
    library_version,
    source_label,
    workload_start_timestamp_label,
)
from aperture_sdk.flow import Flow
from aperture_sdk.utils import TWrappedReturn, run_fn
from opentelemetry import baggage, trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import SERVICE_NAME, SERVICE_VERSION, Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.util import types as otel_types

TApertureClient = TypeVar("TApertureClient", bound="ApertureClient")
TWrappedFunction = Callable[..., TWrappedReturn]
Labels = Dict[str, str]


class ApertureClient:
    def __init__(
        self,
        channel: grpc.Channel,
        otlp_exporter: OTLPSpanExporter,
    ):
        self.logger = logging.getLogger("aperture-py-sdk")

        resource = Resource.create(
            {
                SERVICE_NAME: library_name,
                SERVICE_VERSION: library_version,
            }
        )
        tracer_provider = TracerProvider(resource=resource)
        trace.set_tracer_provider(tracer_provider)
        self.tracer = trace.get_tracer(library_name, library_version)

        span_processor = BatchSpanProcessor(otlp_exporter)
        tracer_provider.add_span_processor(span_processor)
        self.otlp_exporter = otlp_exporter
        self.grpc_channel = channel

    @classmethod
    def new_client(
        cls: Type[TApertureClient],
        endpoint: str = "http://localhost:4317",
        insecure: bool = False,
        grpc_timeout: datetime.timedelta = default_grpc_reconnection_time,
        credentials: Optional[grpc.ChannelCredentials] = None,
        compression: grpc.Compression = grpc.Compression.NoCompression,
    ) -> TApertureClient:
        if credentials is None:
            credentials = grpc.ssl_channel_credentials()
        otlp_exporter = OTLPSpanExporter(
            endpoint=endpoint,
            insecure=insecure,
            credentials=credentials,
            compression=compression,
            timeout=int(grpc_timeout.total_seconds()),
        )
        grpc_channel_options_dict = {
            "grpc.keepalive_time_ms": 10000,
            "grpc.keepalive_timeout_ms": 5000,
            "grpc.keepalive_permit_without_calls": 1,
        }
        grpc_channel_options = [(k, v) for k, v in grpc_channel_options_dict.items()]
        grpc_channel = (
            grpc.insecure_channel(
                endpoint, compression=compression, options=grpc_channel_options
            )
            if insecure
            else grpc.secure_channel(
                endpoint,
                credentials,
                compression=compression,
                options=grpc_channel_options,
            )
        )
        return cls(
            channel=grpc_channel,
            otlp_exporter=otlp_exporter,
        )

    def start_flow(
        self,
        control_point: str,
        explicit_labels: Optional[Labels] = None,
        check_timeout: datetime.timedelta = default_rpc_timeout,
        ramp_mode: bool = False,
    ) -> Flow:
        labels: Labels = {}
        labels.update({key: str(value) for key, value in baggage.get_all().items()})
        # Explicit labels override baggage
        labels.update(explicit_labels or {})
        request = CheckRequest(
            control_point=control_point, labels=labels, ramp_mode=ramp_mode
        )
        span_attributes: otel_types.Attributes = {
            flow_start_timestamp_label: time.monotonic_ns(),
            source_label: "sdk",
        }

        span = self.tracer.start_span("Aperture Check", attributes=span_attributes)
        stub = FlowControlServiceStub(self.grpc_channel)
        try:
            # stub.Check is typed to accept an int, but it actually accepts a float
            timeout = typing.cast(int, check_timeout.total_seconds())
            response = (
                stub.Check(request)
                if timeout == 0
                else stub.Check(request, timeout=timeout)
            )
        except grpc.RpcError as e:
            self.logger.debug(f"Aperture gRPC call failed: {e.details()}")
            response = None
        span.set_attribute(workload_start_timestamp_label, time.monotonic_ns())
        return Flow(span=span, check_response=response)

    def decorate(
        self,
        control_point: str,
        explicit_labels: Optional[Dict[str, str]] = None,
        on_reject: Optional[Callable] = None,
        check_timeout: datetime.timedelta = default_rpc_timeout,
        ramp_mode: bool = False,
    ) -> Callable[[TWrappedFunction], TWrappedFunction]:
        def decorator(fn: TWrappedFunction) -> TWrappedFunction:
            @functools.wraps(fn)
            async def wrapper(*args, **kwargs):
                with self.start_flow(
                    control_point, explicit_labels, check_timeout, ramp_mode
                ) as flow:
                    if flow.should_run():
                        return await run_fn(fn, *args, **kwargs)
                    else:
                        if on_reject:
                            return on_reject()
                        raise RejectedFlowException("Flow was rejected")

            return wrapper

        return decorator

    def close(self):
        self.otlp_exporter.shutdown()


class ApertureException(Exception):
    pass


class RejectedFlowException(ApertureException):
    pass
